@@ -11,20 +11,21 @@
 
 import asyncio
 import time
+from asyncio import sleep
 
-from aiorpcx import TaskGroup, run_in_thread, CancelledError
+from aiorpcx import TaskGroup, CancelledError
 
 import electrumx
 from electrumx.server.daemon import DaemonError
 from electrumx.lib.hash import hash_to_hex_str, HASHX_LEN
 from electrumx.lib.script import is_unspendable_legacy, is_unspendable_genesis
 from electrumx.lib.util import (
-    chunks, class_logger, pack_le_uint32, pack_le_uint64, unpack_le_uint64
+    class_logger, pack_le_uint32, pack_le_uint64, unpack_le_uint64
 )
 from electrumx.server.db import FlushData
 
 
-class Prefetcher(object):
+class Prefetcher:
     '''Prefetches blocks (in the forward direction only).'''
 
     def __init__(self, daemon, coin, blocks_event):
@@ -54,7 +55,7 @@ class Prefetcher(object):
                 # Sleep a while if there is nothing to prefetch
                 await self.refill_event.wait()
                 if not await self._prefetch_blocks():
-                    await asyncio.sleep(self.polling_delay)
+                    await sleep(self.polling_delay)
             except DaemonError as e:
                 self.logger.info(f'ignoring daemon error: {e}')
             except CancelledError as e:
@@ -147,7 +148,7 @@ class ChainError(Exception):
     '''Raised on error processing blocks.'''
 
 
-class BlockProcessor(object):
+class BlockProcessor:
     '''Process blocks and update the DB state to match.
 
     Employ a prefetcher to prefetch blocks in batches for processing.
@@ -160,15 +161,25 @@ class BlockProcessor(object):
         self.daemon = daemon
         self.notifications = notifications
 
-        self.coin = env.coin
+        # Set when there is block processing to do, e.g. when new blocks come in, or a
+        # reorg is needed.
         self.blocks_event = asyncio.Event()
+
+        # If the lock is successfully acquired, in-memory chain state
+        # is consistent with self.height
+        self.state_lock = asyncio.Lock()
+
+        # Signalled after backing up during a reorg
+        self.backed_up_event = asyncio.Event()
+
+        self.coin = env.coin
         self.prefetcher = Prefetcher(daemon, env.coin, self.blocks_event)
         self.logger = class_logger(__name__, self.__class__.__name__)
 
         # Meta
         self.next_cache_check = 0
         self.touched = set()
-        self.reorg_count = 0
+        self.reorg_count = None
         self.height = -1
         self.tip = None
         self.tx_count = 0
@@ -183,107 +194,65 @@ class BlockProcessor(object):
         self.utxo_cache = {}
         self.db_deletes = []
 
-        # If the lock is successfully acquired, in-memory chain state
-        # is consistent with self.height
-        self.state_lock = asyncio.Lock()
-
-        # Signalled after backing up during a reorg
-        self.backed_up_event = asyncio.Event()
-
-    async def run_in_thread_with_lock(self, func, *args):
-        # Run in a thread to prevent blocking.  Shielded so that
-        # cancellations from shutdown don't lose work - when the task
-        # completes the data will be flushed and then we shut down.
-        # Take the state lock to be certain in-memory state is
-        # consistent and not being updated elsewhere.
-        async def run_in_thread_locked():
+    async def run_with_lock(self, coro):
+        # Shielded so that cancellations from shutdown don't lose work.  Cancellation will
+        # cause fetch_and_process_blocks to block on the lock in flush(), the task completes,
+        # and then the data is flushed.  We also don't want user-signalled reorgs to happen
+        # in the middle of processing blocks; they need to wait.
+        async def run_locked():
             async with self.state_lock:
-                return await run_in_thread(func, *args)
-        return await asyncio.shield(run_in_thread_locked())
+                return await coro
+        return await asyncio.shield(run_locked())
 
-    async def check_and_advance_blocks(self, raw_blocks):
-        '''Process the list of raw blocks passed.  Detects and handles
-        reorgs.
-        '''
-        if not raw_blocks:
-            return
-        first = self.height + 1
-        blocks = [self.coin.block(raw_block, first + n)
-                  for n, raw_block in enumerate(raw_blocks)]
-        headers = [block.header for block in blocks]
-        hprevs = [self.coin.header_prevhash(h) for h in headers]
-        chain = [self.tip] + [self.coin.header_hash(h) for h in headers[:-1]]
+    def schedule_reorg(self, count):
+        '''A count >= 0 is a user-forced reorg; < 0 is a natural reorg.'''
+        self.reorg_count = count
+        self.blocks_event.set()
 
-        if hprevs == chain:
-            start = time.monotonic()
-            await self.run_in_thread_with_lock(self.advance_blocks, blocks)
-            await self._maybe_flush()
-            if not self.db.first_sync:
-                s = '' if len(blocks) == 1 else 's'
-                blocks_size = sum(len(block) for block in raw_blocks) / 1_000_000
-                self.logger.info(f'processed {len(blocks):,d} block{s} size {blocks_size:.2f} MB '
-                                 f'in {time.monotonic() - start:.1f}s')
-            if self._caught_up_event.is_set():
-                await self.notifications.on_block(self.touched, self.height)
-            self.touched = set()
-        elif hprevs[0] != chain[0]:
-            await self.reorg_chain()
-        else:
-            # It is probably possible but extremely rare that what
-            # bitcoind returns doesn't form a chain because it
-            # reorg-ed the chain as it was processing the batched
-            # block hash requests.  Should this happen it's simplest
-            # just to reset the prefetcher and try again.
-            self.logger.warning('daemon blocks do not form a chain; '
-                                'resetting the prefetcher')
-            await self.prefetcher.reset_height(self.height)
-
-    async def reorg_chain(self, count=None):
+    async def _reorg_chain(self, count):
         '''Handle a chain reorganisation.
 
         Count is the number of blocks to simulate a reorg, or None for
         a real reorg.'''
-        if count is None:
+        if count < 0:
             self.logger.info('chain reorg detected')
         else:
             self.logger.info(f'faking a reorg of {count:,d} blocks')
         await self.flush(True)
 
-        async def get_raw_blocks(last_height, hex_hashes):
-            heights = range(last_height, last_height - len(hex_hashes), -1)
+        async def get_raw_block(hex_hash, height):
             try:
-                blocks = [self.db.read_raw_block(height) for height in heights]
-                self.logger.info(f'read {len(blocks)} blocks from disk')
-                return blocks
+                block = self.db.read_raw_block(height)
+                self.logger.info(f'read block {hex_hash} at height {height:,d} from disk')
             except FileNotFoundError:
-                return await self.daemon.raw_blocks(hex_hashes)
+                block = await self.daemon.raw_blocks([hex_hash])[0]
+                self.logger.info(f'obtained block {hex_hash} at height {height:,d} from daemon')
+            return block
 
-        def flush_backup():
-            # self.touched can include other addresses which is
-            # harmless, but remove None.
+        _start, height, hashes = await self._reorg_hashes(count)
+        hex_hashes = [hash_to_hex_str(block_hash) for block_hash in hashes]
+        for hex_hash in reversed(hex_hashes):
+            raw_block = await get_raw_block(hex_hash, height)
+            await self._backup_block(raw_block)
+            # self.touched can include other addresses which is harmless, but remove None.
             self.touched.discard(None)
             self.db.flush_backup(self.flush_data(), self.touched)
+            height -= 1
+
+        self.logger.info('backed up to height {:,d}'.format(self.height))
 
-        _start, last, hashes = await self.reorg_hashes(count)
-        # Reverse and convert to hex strings.
-        hashes = [hash_to_hex_str(hash) for hash in reversed(hashes)]
-        for hex_hashes in chunks(hashes, 50):
-            raw_blocks = await get_raw_blocks(last, hex_hashes)
-            await self.run_in_thread_with_lock(self.backup_blocks, raw_blocks)
-            await self.run_in_thread_with_lock(flush_backup)
-            last -= len(raw_blocks)
         await self.prefetcher.reset_height(self.height)
         self.backed_up_event.set()
         self.backed_up_event.clear()
 
-    async def reorg_hashes(self, count):
+    async def _reorg_hashes(self, count):
         '''Return a pair (start, last, hashes) of blocks to back up during a
         reorg.
 
         The hashes are returned in order of increasing height.  Start
         is the height of the first hash, last of the last.
         '''
-        start, count = await self.calc_reorg_range(count)
+        start, count = await self._calc_reorg_range(count)
         last = start + count - 1
         s = '' if count == 1 else 's'
         self.logger.info(f'chain was reorganised replacing {count:,d} '
@@ -291,7 +260,7 @@ class BlockProcessor(object):
 
         return start, last, await self.db.fs_block_hashes(start, count)
 
-    async def calc_reorg_range(self, count):
+    async def _calc_reorg_range(self, count):
         '''Calculate the reorg range'''
 
         def diff_pos(hashes1, hashes2):
@@ -302,7 +271,7 @@ class BlockProcessor(object):
                     return n
             return len(hashes)
 
-        if count is None:
+        if count < 0:
             # A real reorg
             start = self.height - 1
             count = 1
@@ -342,21 +311,8 @@ class BlockProcessor(object):
                          self.db_deletes, self.tip)
 
     async def flush(self, flush_utxos):
-        def flush():
-            self.db.flush_dbs(self.flush_data(), flush_utxos,
-                              self.estimate_txs_remaining)
-        await self.run_in_thread_with_lock(flush)
-
-    async def _maybe_flush(self):
-        # If caught up, flush everything as client queries are
-        # performed on the DB.
-        if self._caught_up_event.is_set():
-            await self.flush(True)
-        elif time.monotonic() > self.next_cache_check:
-            flush_arg = self.check_cache_size()
-            if flush_arg is not None:
-                await self.flush(flush_arg)
-            self.next_cache_check = time.monotonic() + 30
+        self.db.flush_dbs(self.flush_data(), flush_utxos, self.estimate_txs_remaining)
+        self.next_cache_check = time.monotonic() + 30
 
     def check_cache_size(self):
         '''Flush a cache if it gets too big.'''
@@ -384,28 +340,55 @@ class BlockProcessor(object):
             return utxo_MB >= cache_MB * 4 // 5
         return None
 
-    def advance_blocks(self, blocks):
-        '''Synchronously advance the blocks.
+    async def _advance_blocks(self, raw_blocks):
+        '''Process the list of raw blocks passed.  Detects and handles reorgs.'''
+        start = time.monotonic()
+        for raw_block in raw_blocks:
+            block = self.coin.block(raw_block)
+            if self.coin.header_prevhash(block.header) != self.tip:
+                self.schedule_reorg(-1)
+                return
+            await self._advance_block(block)
+        end = time.monotonic()
 
-        It is already verified they correctly connect onto our tip.
-        '''
+        if not self.db.first_sync:
+            s = '' if len(raw_blocks) == 1 else 's'
+            blocks_size = sum(len(block) for block in raw_blocks) / 1_000_000
+            self.logger.info(f'processed {len(raw_blocks):,d} block{s} size {blocks_size:.2f} MB '
+                             f'in {end - start:.1f}s')
+
+        # If caught up, flush everything as client queries are performed on the DB,
+        # otherwise check at regular intervals.
+        if self.height == self.daemon.cached_height():
+            await self.flush(True)
+            await self._on_caught_up()
+        elif end > self.next_cache_check:
+            flush_arg = self.check_cache_size()
+            if flush_arg is not None:
+                await self.flush(flush_arg)
+
+        if self._caught_up_event.is_set():
+            await self.notifications.on_block(self.touched, self.height)
+
+        self.touched = set()
+
+    async def _advance_block(self, block):
+        '''Advance once block.  It is already verified they correctly connect onto our tip.'''
         min_height = self.db.min_undo_height(self.daemon.cached_height())
-        height = self.height
-        genesis_activation = self.coin.GENESIS_ACTIVATION
+        height = self.height + 1
 
-        for block in blocks:
-            height += 1
-            is_unspendable = (is_unspendable_genesis if height >= genesis_activation
-                              else is_unspendable_legacy)
-            undo_info = self.advance_txs(block.transactions, is_unspendable)
-            if height >= min_height:
-                self.undo_infos.append((undo_info, height))
-                self.db.write_raw_block(block.raw, height)
+        is_unspendable = (is_unspendable_genesis if height >= self.coin.GENESIS_ACTIVATION
+                          else is_unspendable_legacy)
+        undo_info = self.advance_txs(block.transactions, is_unspendable)
+        if height >= min_height:
+            self.undo_infos.append((undo_info, height))
+            self.db.write_raw_block(block.raw, height)
 
-        headers = [block.header for block in blocks]
         self.height = height
-        self.headers.extend(headers)
-        self.tip = self.coin.header_hash(headers[-1])
+        self.headers.append(block.header)
+        self.tip = self.coin.header_hash(block.header)
+
+        await sleep(0)
 
     def advance_txs(self, txs, is_unspendable):
         self.tx_hashes.append(b''.join(tx_hash for tx, tx_hash in txs))
@@ -459,36 +442,36 @@ class BlockProcessor(object):
 
         return undo_info
 
-    def backup_blocks(self, raw_blocks):
-        '''Backup the raw blocks and flush.
+    async def _backup_block(self, raw_block):
+        '''Backup the raw block and flush.
 
-        The blocks should be in order of decreasing height, starting at.
-        self.height.  A flush is performed once the blocks are backed up.
+        The blocks should be in order of decreasing height, starting at.  self.height.  A
+        flush is performed once the blocks are backed up.
         '''
         self.db.assert_flushed(self.flush_data())
-        assert self.height >= len(raw_blocks)
+        assert self.height > 0
         genesis_activation = self.coin.GENESIS_ACTIVATION
 
         coin = self.coin
-        for raw_block in raw_blocks:
-            # Check and update self.tip
-            block = coin.block(raw_block, self.height)
-            header_hash = coin.header_hash(block.header)
-            if header_hash != self.tip:
-                raise ChainError('backup block {} not tip {} at height {:,d}'
-                                 .format(hash_to_hex_str(header_hash),
-                                         hash_to_hex_str(self.tip),
-                                         self.height))
-            self.tip = coin.header_prevhash(block.header)
-            is_unspendable = (is_unspendable_genesis if self.height >= genesis_activation
-                              else is_unspendable_legacy)
-            self.backup_txs(block.transactions, is_unspendable)
-            self.height -= 1
-            self.db.tx_counts.pop()
 
-        self.logger.info('backed up to height {:,d}'.format(self.height))
+        # Check and update self.tip
+        block = coin.block(raw_block)
+        header_hash = coin.header_hash(block.header)
+        if header_hash != self.tip:
+            raise ChainError('backup block {} not tip {} at height {:,d}'
+                             .format(hash_to_hex_str(header_hash),
+                                     hash_to_hex_str(self.tip),
+                                     self.height))
+        self.tip = coin.header_prevhash(block.header)
+        is_unspendable = (is_unspendable_genesis if self.height >= genesis_activation
+                          else is_unspendable_legacy)
+        self._backup_txs(block.transactions, is_unspendable)
+        self.height -= 1
+        self.db.tx_counts.pop()
 
-    def backup_txs(self, txs, is_unspendable):
+        await sleep(0)
+
+    def _backup_txs(self, txs, is_unspendable):
         # Prevout values, in order down the block (coinbase first if present)
         # undo_info is in reverse block order
         undo_info = self.db.read_undo_info(self.height)
@@ -623,33 +606,38 @@ class BlockProcessor(object):
         raise ChainError('UTXO {} / {:,d} not found in "h" table'
                          .format(hash_to_hex_str(tx_hash), tx_idx))
 
-    async def _process_prefetched_blocks(self):
+    async def _process_blocks(self):
         '''Loop forever processing blocks as they arrive.'''
-        while True:
-            if self.height == self.daemon.cached_height():
-                if not self._caught_up_event.is_set():
-                    await self._first_caught_up()
-                    self._caught_up_event.set()
-            await self.blocks_event.wait()
-            self.blocks_event.clear()
-            if self.reorg_count:
-                await self.reorg_chain(self.reorg_count)
-                self.reorg_count = 0
+        async def process_event():
+            '''Perform a pending reorg or process prefetched blocks.'''
+            if self.reorg_count is not None:
+                await self._reorg_chain(self.reorg_count)
+                self.reorg_count = None
+                # Prefetcher block cache cleared so nothing to process
             else:
                 blocks = self.prefetcher.get_prefetched_blocks()
-                await self.check_and_advance_blocks(blocks)
+                await self._advance_blocks(blocks)
 
-    async def _first_caught_up(self):
-        self.logger.info(f'caught up to height {self.height}')
-        # Flush everything but with first_sync->False state.
-        first_sync = self.db.first_sync
-        self.db.first_sync = False
-        await self.flush(True)
-        if first_sync:
-            self.logger.info(f'{electrumx.version} synced to '
-                             f'height {self.height:,d}')
-        # Reopen for serving
-        await self.db.open_for_serving()
+        # This must be done to set state before the main loop
+        if self.height == self.daemon.cached_height():
+            await self._on_caught_up()
+
+        while True:
+            await self.blocks_event.wait()
+            self.blocks_event.clear()
+            await self.run_with_lock(process_event())
+
+    async def _on_caught_up(self):
+        if not self._caught_up_event.is_set():
+            self._caught_up_event.set()
+            self.logger.info(f'caught up to height {self.height}')
+            # Flush everything but with first_sync->False state.
+            first_sync = self.db.first_sync
+            self.db.first_sync = False
+            if first_sync:
+                self.logger.info(f'{electrumx.version} synced to height {self.height:,d}')
+            # Reopen for serving
+            await self.db.open_for_serving()
 
     async def _first_open_dbs(self):
         await self.db.open_for_sync()
@@ -676,12 +664,14 @@ class BlockProcessor(object):
         try:
             async with TaskGroup() as group:
                 await group.spawn(self.prefetcher.main_loop(self.height))
-                await group.spawn(self._process_prefetched_blocks())
+                await group.spawn(self._process_blocks())
+            raise group.exception
         # Don't flush for arbitrary exceptions as they might be a cause or consequence of
         # corrupted data
         except CancelledError:
             self.logger.info('flushing to DB for a clean shutdown...')
-            await self.flush(True)
+            await self.run_with_lock(self.flush(True))
+            self.logger.info('flushed cleanly')
 
     def force_chain_reorg(self, count):
         '''Force a reorg of the given number of blocks.
@@ -689,145 +679,6 @@ class BlockProcessor(object):
         Returns True if a reorg is queued, false if not caught up.
         '''
         if self._caught_up_event.is_set():
-            self.reorg_count = count
-            self.blocks_event.set()
+            self.schedule_reorg(count)
             return True
         return False
-
-
-class DecredBlockProcessor(BlockProcessor):
-    async def calc_reorg_range(self, count):
-        start, count = await super().calc_reorg_range(count)
-        if start > 0:
-            # A reorg in Decred can invalidate the previous block
-            start -= 1
-            count += 1
-        return start, count
-
-
-class NameIndexBlockProcessor(BlockProcessor):
-
-    def advance_txs(self, txs, is_unspendable):
-        result = super().advance_txs(txs, is_unspendable)
-
-        tx_num = self.tx_count - len(txs)
-        script_name_hashX = self.coin.name_hashX_from_script
-        update_touched = self.touched.update
-        hashXs_by_tx = []
-        append_hashXs = hashXs_by_tx.append
-
-        for tx, _tx_hash in txs:
-            hashXs = []
-            append_hashX = hashXs.append
-
-            # Add the new UTXOs and associate them with the name script
-            for txout in tx.outputs:
-                # Get the hashX of the name script.  Ignore non-name scripts.
-                hashX = script_name_hashX(txout.pk_script)
-                if hashX:
-                    append_hashX(hashX)
-
-            append_hashXs(hashXs)
-            update_touched(hashXs)
-            tx_num += 1
-
-        self.db.history.add_unflushed(hashXs_by_tx, self.tx_count - len(txs))
-
-        return result
-
-
-class LTORBlockProcessor(BlockProcessor):
-
-    def advance_txs(self, txs, is_unspendable):
-        self.tx_hashes.append(b''.join(tx_hash for tx, tx_hash in txs))
-
-        # Use local vars for speed in the loops
-        undo_info = []
-        tx_num = self.tx_count
-        script_hashX = self.coin.hashX_from_script
-        put_utxo = self.utxo_cache.__setitem__
-        spend_utxo = self.spend_utxo
-        undo_info_append = undo_info.append
-        update_touched = self.touched.update
-        to_le_uint32 = pack_le_uint32
-        to_le_uint64 = pack_le_uint64
-
-        hashXs_by_tx = [set() for _ in txs]
-
-        # Add the new UTXOs
-        for (tx, tx_hash), hashXs in zip(txs, hashXs_by_tx):
-            add_hashXs = hashXs.add
-            tx_numb = to_le_uint64(tx_num)[:5]
-
-            for idx, txout in enumerate(tx.outputs):
-                # Ignore unspendable outputs
-                if is_unspendable(txout.pk_script):
-                    continue
-
-                # Get the hashX
-                hashX = script_hashX(txout.pk_script)
-                add_hashXs(hashX)
-                put_utxo(tx_hash + to_le_uint32(idx),
-                         hashX + tx_numb + to_le_uint64(txout.value))
-            tx_num += 1
-
-        # Spend the inputs
-        # A separate for-loop here allows any tx ordering in block.
-        for (tx, tx_hash), hashXs in zip(txs, hashXs_by_tx):
-            add_hashXs = hashXs.add
-            for txin in tx.inputs:
-                if txin.is_generation():
-                    continue
-                cache_value = spend_utxo(txin.prev_hash, txin.prev_idx)
-                undo_info_append(cache_value)
-                add_hashXs(cache_value[:-13])
-
-        # Update touched set for notifications
-        for hashXs in hashXs_by_tx:
-            update_touched(hashXs)
-
-        self.db.history.add_unflushed(hashXs_by_tx, self.tx_count)
-
-        self.tx_count = tx_num
-        self.db.tx_counts.append(tx_num)
-
-        return undo_info
-
-    def backup_txs(self, txs, is_unspendable):
-        undo_info = self.db.read_undo_info(self.height)
-        if undo_info is None:
-            raise ChainError('no undo information found for height {:,d}'
-                             .format(self.height))
-
-        # Use local vars for speed in the loops
-        put_utxo = self.utxo_cache.__setitem__
-        spend_utxo = self.spend_utxo
-        add_touched = self.touched.add
-        undo_entry_len = 13 + HASHX_LEN
-
-        # Restore coins that had been spent
-        # (may include coins made then spent in this block)
-        n = 0
-        for tx, tx_hash in txs:
-            for txin in tx.inputs:
-                if txin.is_generation():
-                    continue
-                undo_item = undo_info[n:n + undo_entry_len]
-                put_utxo(txin.prev_hash + pack_le_uint32(txin.prev_idx), undo_item)
-                add_touched(undo_item[:-13])
-                n += undo_entry_len
-
-        assert n == len(undo_info)
-
-        # Remove tx outputs made in this block, by spending them.
-        for tx, tx_hash in txs:
-            for idx, txout in enumerate(tx.outputs):
-                # Spend the TX outputs.  Be careful with unspendable
-                # outputs - we didn't save those in the first place.
-                if is_unspendable(txout.pk_script):
-                    continue
-
-                cache_value = spend_utxo(tx_hash, idx)
-                add_touched(cache_value[:-13])
-
-        self.tx_count -= len(txs)
