@@ -258,9 +258,8 @@ class PeerManager:
                 self.logger.info(f'{peer_text} {e}')
             except CancelledError:
                 # A send_request was cancelled
-                self.logger.error('peer dropped verification connection')
-                peer.mark_bad()
-                break
+                self.logger.error(f'{peer_text} dropped verification connection')
+                raise
 
         if is_good:
             # Monotonic time would be better, but last_good and last_try are
@@ -347,6 +346,9 @@ class PeerManager:
             peers_task = await g.spawn(self._send_peers_subscribe
                                        (session, peer))
 
+        # Propagate failed task exception
+        g.results    # pylint:disable=W0104
+
         # Process reported peers if remote peer is good
         peers = peers_task.result()
         await self._note_peers(peers)
