@@ -148,7 +148,7 @@ class SessionManager:
         # Set up the RPC request handlers
         cmds = ('add_peer daemon_url disconnect getinfo groups log peers '
                 'query reorg sessions stop'.split())
-        LocalRPC.request_handlers = {cmd: getattr(self, 'rpc_' + cmd)
+        self.rpc_request_handlers = {cmd: getattr(self, 'rpc_' + cmd)
                                      for cmd in cmds}
 
     def _ssl_context(self):
@@ -621,10 +621,13 @@ class SessionManager:
                 await group.spawn(self._manage_servers())
         finally:
             # Close servers then sessions
+            self.logger.info('stopping servers')
             await self._stop_servers(self.servers.keys())
+            self.logger.info('closing connections...')
             async with TaskGroup() as group:
                 for session in list(self.sessions):
                     await group.spawn(session.close(force_after=1))
+            self.logger.info('connections closed')
 
     def extra_cost(self, session):
         # Note there is no guarantee that session is still in self.sessions.  Example traceback:
@@ -1330,14 +1333,6 @@ class ElectrumX(SessionBase):
                            f'network rules.\n\n{message}\n[{raw_tx}]') from None
         else:
             self.txs_sent += 1
-            client_ver = util.protocol_tuple(self.client)
-            if client_ver != (0, ):
-                msg = self.coin.warn_old_client_on_tx_broadcast(client_ver)
-                if msg:
-                    self.logger.info(f'sent tx: {hex_hash}. and warned user to upgrade their '
-                                     f'client from {self.client}')
-                    return msg
-
             self.logger.info(f'sent tx: {hex_hash}')
             return hex_hash
 
@@ -1440,6 +1435,7 @@ class LocalRPC(SessionBase):
         super().__init__(*args, **kwargs)
         self.client = 'RPC'
         self.connection.max_response_size = 0
+        self.request_handlers = self.session_mgr.rpc_request_handlers
 
     def protocol_version_string(self):
         return 'RPC'
